/*
 * Copyright (c) 2018-2021 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Mar 12 2018
 */
#ifndef H_LIBMRSIM
#define H_LIBMRSIM

#include <sys/types.h>
#include <stdint.h>
#include <time.h>

/* simulation parameters */
struct mrsim_param_s {
  /* motor constants */
  struct {
    double J; /* inertia, kg.m^2 */
    double b; /* drag, N.m.s */
    double K; /* torque constant, N.m/A */
    double R; /* electric resistance, Ohm */
    double L; /* electric inductance, H */
    double V; /* battery voltage, V */
    double I; /* maximum current, A */
  } motor;

  /* dynamics */
  struct {
    const uint8_t rotors;/* number of rotors */
    double * const G;	/* mapping from propellers velocity^2 to wrench (6×n) */
    double J[3 * 3];	/* inertia matrix */
    double m;		/* mass */
    double g;		/* gravity */
    double B[3];	/* magnetic field, NWU frame */
  } body;

  /* state */
  struct {
    double pstddev, qstddev, vstddev, wstddev, astddev, Bstddev;
    double imup, magp, batp, motp;
  } state;
};


/* motor state */
struct mrsimstate_motor_s {
  bool emerg, spinning, starting, disabled, reverse;

  double spin; /* +1.0/-1.0 */
  double pwm;
  double wd, w, i;
};


/* body state */
struct mrsimstate_wrench_s {
  double thrust[3], torque[3];	/* in body frame */
};

struct mrsimstate_body_s {
  double px, py, pz;		/* position */
  double qw, qx, qy, qz;	/* orientation */

  double vx, vy, vz;		/* linear velocity in world frame */
  double wx, wy, wz;		/* angular velocity in world frame */

  double ax, ay, az;		/* linear acceleration in world frame */
  double dwx, dwy, dwz;		/* angular acceleration in world frame */
};


/* simulation context */
struct mrsim_s;


#ifdef __cplusplus
extern "C" {
#endif
  struct mrsim_s *	mrsim_init(uint8_t rotors);
  int			mrsim_open_pty(const struct mrsim_s *mrsim,
                                const char *device);
  double		mrsim_gnoise(double stddev);
  void			mrsim_fini(struct mrsim_s *mrsim);

  int			mrsim_wait_msg(const struct mrsim_s *mrsim,
                                int timeoutms);
  int			mrsim_process_msgs(const struct mrsim_s *mrsim);
  int			mrsim_send_msgs(const struct mrsim_s *mrsim);

  void			mrsim_update_time(const struct mrsim_s *mrsim,
                                struct timespec now);
  void			mrsim_update_motors(const struct mrsim_s *mrsim);
  void			mrsim_update_wrench(const struct mrsim_s *mrsim);
  void			mrsim_update_body(const struct mrsim_s *mrsim);
  void			mrsim_update_imu(const struct mrsim_s *mrsim);

  struct mrsim_param_s *	mrsim_params(const struct mrsim_s *mrsim);
  struct mrsimstate_motor_s *	mrsim_motor(const struct mrsim_s *mrsim,
                                        uint8_t index);
  struct mrsimstate_wrench_s *	mrsim_wrench(const struct mrsim_s *mrsim);
  struct mrsimstate_body_s *	mrsim_body(const struct mrsim_s *mrsim);

#ifdef __cplusplus
}
#endif


#endif /* H_LIBMRSIM */
